{***********************************************************************\ 
*									* 
*   File: scorpion/include/Pascal/reader.h 
*				 					* 
*   Copyright (C) 1991 Michael Shapiro and Richard Snodgrass
*									* 
*   The Scorpion System is free software in the public domain; you can  * 
*   redistribute it and/or modify it as you wish. We ask that you 	* 
*   retain credits referencing the University of Arizona and that you	* 
*   identify any changes you make.					* 
*									* 
*   Report problems to scorpion-project@cs.arizona.edu			* 
*   Direct all inquiries to:	The Scorpion Project			* 
*				Department of Computer Science		* 
*				Gould-Simpson Building			* 
*				University of Arizona			* 
*				Tucson, AZ 85721			* 
*				U.S.A.					* 
*									* 
*   Revision Log:							* 
*	$Log:$ 
*									* 
*   Edit Log:								* 
*									* 
\***********************************************************************} 

typedef enum {
LexCloseAngleBracket, LexCloseBrace, LexCloseSquareBracket, LexEOF, LexFalse,
LexTrue, LexInteger, LexLabelDef, LexLabelRef, LexName, LexOpenAngleBracket,
LexOpenBrace, LexOpenSquareBracket, LexRational, LexSemiColon, LexString }
LexType;

typedef struct {
	int line;
} PosType;

typedef int AttrDesc;

typedef enum {SeqLabel, NodeLabel} LabelKind;
typedef struct LabelRecord *LabelType;
typedef struct LabelRefRecord *LabelRefType;
typedef struct RecRepType *RepType;
struct LabelRefRecord {
	LabelRefType next;	/* for collision resolution by chaining */
	LabelType RefList;
	int Unresolved;
	int Generated;
	String Name;
	RepType Val;
};
struct LabelRecord
{	LabelType next;
	LabelRefType parent;
	PosType Pos;
	LabelKind Kind;
	union
	{	RepType *SLabel;	/* pointer to where to put it */
		struct
		{	NodeType NodeField;
			AttrDesc Desc;
		} NLabel;
	} UnionField;
};

typedef struct {
	LexType Kind;
	PosType Pos;
	String Token;
} Lexeme;

typedef struct CellType *SeqType;

typedef enum { 
RepLabelRef, RepNode, AtrInteger, AtrRational, AtrString, AtrBoolean,
AtrSequence, RepPrivate
} RepKind;

struct RecRepType {
	PosType Pos;
	String StringField;
	RepKind Kind;
	int Labelled;
	union {
		SeqType seqVal;
		int boolVal;
		LabelRefType LabelField;
		NodeType NodeField;
		int PrivateField;
	}	UnionField;
};

struct CellType {
	SeqType next;
	RepType value;
};

typedef enum {
	ErrAngleExpected, ErrBraceExpected, ErrEOF, ErrSemiColonExpected,
	ErrSkipStopped, ErrValueExpected, ErrBadNodeType, ErrNodeExpected,
	ErrEOFExpected, ErrBadRootType, ErrNoRoot, ErrBadLex, ErrBadAttribute,
	ErrBadClassNode, ErrBadEscapedChar }
ErrType;

typedef struct {
	char *nodetypename;
	int nodetypeconstant;
} ntype;

#define MAXATTRIBUTES 100
typedef struct {
	char *attributes[MAXATTRIBUTES];
} natype;
NodeType Reader ();

/* error functions */
void ErrInit();
void ErrFinish();

/* sequence functions */

SeqType AddSeq();

/* functions to create RepTypes */

RepType MakeVal();

RepType MakeBool();

RepType MakeNode();

RepType MakeSeq();

/* is the input root in the root set for the port? */

int InRootSet();

/* attribute table functions */

AttrDesc FindAttribute();

int FindNodeType();

/* lexical functions */

Lexeme GetLex();

void LexInit();

/* label functions */

RepType FakeLabel();

void DelaySeqRef();

RepType LabelRef();

void SaveLabel();

void DelayedRef();

void LabelInit();

int LabelFinish();

/* process-specific functions */

void AttrStore();

NodeType NodeAlloc();

/* misc. reader funcs */

String GString();

void DelRep();

extern int OK;
extern Lexeme Lex;
extern void GAInteger(), GSetInteger(), GAfloat(), GAdouble();
extern LInteger GLInteger();
extern Lfloat GLfloat();
extern Ldouble GLdouble();
extern LString GLString();
extern LBoolean GLBoolean();
# define GSetFloat(X) GLfloat(X)
# define GSetDouble(X) GLdouble(X)
# define GSetString(X) GLString(X)
# define Gint(R) (atoi((R)->StringField))
# define Guint(R) Gint(R)
# define Gushort(R) Gint(R)
# define Gshort(R) Gint(R)
# define Gchar(R) Gint(R)
# define Guchar(R) Gint(R)
# define Glong(R) Gint(R)
# define Gulong(R) Gint(R)
# define Gfloat(R) (atof((R)->StringField))
# define Gdouble(R) Gfloat(R)
extern int atoi();
extern double atof();
